package nts.parser;

import java.util.*;

import nts.interf.*;
import nts.interf.base.IExpr;
import nts.interf.base.IVisitor;

import org.antlr.runtime.Token;
// no instance number specification and only one subsystem s --> s[1]
// no instance number specification and there is subsystem main --> main[1]
public class NTS extends AnnotatedBase implements INTS {
	
	public void accept(IVisitor v) { v.visit(this); }
	
	private Token name;
	public String name() { return name.getText(); }
		
	private static final String MAIN = "main";
	
	private VarTable vt; // global symbol table
	private Expr pre; // precondition on global variables
	private Map<String,Subsystem> sub = new HashMap<String,Subsystem>();
	
	private Map<Token, Expr> inst_tok = new HashMap<Token, Expr>(); // auxiliary mapping subsystem-name-token --> instance-number-expression
	private Map<String, Expr> inst;
	
	// interface
	public VarTable varTable() { return vt; }
	public Expr precondition() { return pre; }
	public List<ISubsystem> subsystems() {
		List<ISubsystem> ret = new ArrayList<ISubsystem>(sub.size());
		int i=0;
		for (Subsystem s : sub.values()) {
			ret.add(i++,s);
		}
		return ret; 
	}
	public Map<String,IExpr> instances() {
		Map<String,IExpr> ret = new HashMap<String,IExpr>();
		for (Map.Entry<String,Expr> e : inst.entrySet()) {
			ret.put(e.getKey(), e.getValue());
		}
		return ret; 
	}
	
	
	public NTS(VarTable aSt, Token aT) {
		vt = aSt;
		name = aT;
	}
	
	void addSubsystem(Subsystem aSub) {
		sub.put(aSub.nameToken().getText(), aSub);
	}
	void setPrecondition(Expr aPre) {
		pre = aPre;
	}
	void addInstance(Token aT, Expr aE) {
		inst_tok.put(aT, aE);
	}
	
	// creates the subsystem-name --> expression mapping
	private void checkInstances() {
		inst = new HashMap<String,Expr>();
		if (inst_tok.isEmpty()) {
			String entryName = null;
			if (sub.size() == 1)
				entryName = sub.keySet().iterator().next();
			else if (sub.containsKey(MAIN))
				entryName = MAIN;
			// no entry of the NTS is specified
			if (entryName == null) {
				System.err.println("Semantic error: No entry-point (neither the "+MAIN+" subsystem, nor a number of instances) of the NTS has been specified.");
				System.exit(1);
			// 'main' subsystem will is the entry
			} else {
				Expr aux = new LitInt(Common.parser.dummyNum(1));
				inst.put(entryName, aux);
				sub.get(entryName).setInstances(aux);
			}
		} else {
			for (Map.Entry<Token, Expr> e : inst_tok.entrySet()) {
				inst.put(e.getKey().getText(), e.getValue());
				Subsystem s;
				Token k = e.getKey();
				if ((s = sub.get(k.getText())) == null) {
					System.err.println("Semantic error at "+Common.at(k)+". Subsystem "+k.getText()+" is not defined.");
					System.exit(1);
				} else {
					s.setInstances(e.getValue());
				}
			}
		}
	}
	
	
	void semanticChecks() {
		
		checkInstances();
		
		// semantic checks for declaration
		// including annotations of global variables
		vt.semanticChecks();
		
		// annotations of global NTS
		super.semanticChecks(vt);
		
		if (pre == null){
			pre = new LitBool(Common.parser.dummyTrue(),true);
		}
		
		{
			SemFlags s = SemFlags.createPrecondition();
			pre = pre.semanticChecks(vt, s);
			pre.checkBool();
		}
		
		// instances
		if (inst != null) {
			SemFlags s = SemFlags.createInstancesSpec();
			for (Expr e : inst.values()) {
				e.semanticChecks(vt, s);
				e.checkInt();
			}
		}
		
		// semantic checks for subsystems
		for (Subsystem s : sub.values()) {
			s.semanticChecks(vt, sub);
		}
	}
}
