package nts.parser;

import java.util.*;

import org.antlr.runtime.Token;

import nts.interf.*;
import nts.interf.base.*;
import nts.interf.expr.IAccessBasic;

public class Subsystem extends AnnotatedBase implements ISubsystem {
	
	public void accept(IVisitor v) { v.visit(this); }
	
	private VarTable vt;
	
	private Token nameToken;
	private Expr inst = null;
	
	private Map<String,IControlState> states = new HashMap<String,IControlState>();
	private List<IControlState> sInit = new LinkedList<IControlState>(),
			sFin = new LinkedList<IControlState>(),
			sErr = new LinkedList<IControlState>();
	
	//private List<IVarTableEntry> in, out;
	private List<IVarTableEntry> in = new LinkedList<IVarTableEntry>(), out = new LinkedList<IVarTableEntry>();
	private List<ITransition> tt = new LinkedList<ITransition>();
	
	public Subsystem(VarTable aSt, Token aT) {
		vt = aSt;
		nameToken = aT;
	}
	
	// interface
	public String name() { return nameToken.getText(); }
	public VarTable varTable() { return vt; }
	public Expr instances() { return inst; }
	public List<IControlState> marksInit() { return sInit; }
	public List<IControlState> marksFinal() { return sFin; }
	public List<IControlState> marksError() { return sErr; }
	public List<IVarTableEntry> varIn() { return in; }
	public List<IVarTableEntry> varOut() { return out; }
	public List<ITransition> transitions() { return tt; }
	
	Token nameToken() { return nameToken; }
	
	void setInstances(Expr aInst) { inst = aInst; }
	void addTransition(Transition aT) { tt.add(aT); }
	
	void declareState(Token aT, Annotations aAnnot) {
		String name = aT.getText();
		if (states.get(name) != null) {
			System.err.println("Semantic error at "+Common.at(aT)+": redeclaration of a control state "+name+".");
			System.exit(1);
		}
		ControlState s = getState(aT);
		s.setAnnotations(aAnnot);
	}
	ControlState getState(Token aT) {
		String name = aT.getText();
		ControlState s = (ControlState)states.get(name);
		if (s == null) {
			s = new ControlState(aT);
			states.put(name, s);
		}
		return s;
	}
	private void checkAdd(Token aT, List<IControlState> aL) {
		IControlState s = getState(aT);
		if (aL.contains(s)) {
			System.err.println("Semantic error at "+Common.at(aT)+": multiple marking of a control state "+aT.getText()+".");
			System.exit(1);
		}
		aL.add(s);
	}
	void addInitial(Token aT) { checkAdd(aT,sInit); }
	void addFinal(Token aT) { checkAdd(aT,sFin); }
	void addError(Token aT) { checkAdd(aT,sErr); }
	
	public void setIn(VarTableEntry aVE) {
		in.add(aVE);
	}
	public void setOut(VarTableEntry aVE) {
		out.add(aVE);
	}
	
	private static void checkCallMatch(Subsystem caller, Subsystem callee, Call call) {
		// return values
		List<IAccessBasic> callerRet = call.returnVars();
		List<IVarTableEntry> calleeRet = callee.varOut();
		for (int i=0; i<callerRet.size(); i++) {
			
			AccessBasic a1 = (AccessBasic)callerRet.get(i);
			Type t1 = a1.type();
			//VarTableEntry e2 = (VarTableEntry)callee.vt.get(calleeRet.get(i));
			Type t2 = (Type)calleeRet.get(i).type();
			
			//if (!t1.matchesForAsgn(t2)) {
			if (!t1.matches(t2)) {
				System.err.println("Semantic error at "+Common.at(a1.first)+": type does not match with the callee's return type.");
				System.exit(1);
			}
		}
		
		// actual parameters
		List<IExpr> callerAct = call.actualParameters();
		List<IVarTableEntry> calleeIn = callee.varIn();
		for (int i=0; i<callerAct.size(); i++) {
			
			// note: expressions over NAT have integer type
			// check: 1) type of expression must match 
			
			Expr e1 = (Expr)callerAct.get(i);
			Type t1 = e1.type();
			
			//VarTableEntry e2 = (VarTableEntry)callee.vt.get(calleeIn.get(i));
			Type t2 = (Type)calleeIn.get(i).type();
			
			if (!t1.matches(t2)) {
				System.err.println("Semantic error at "+Common.at(e1.first())+": type of the expression does not match with the type of corresponding formal parameters.");
				System.exit(1);
			}
		}
	}
	
	void semanticChecks(VarTable aVT, Map<String, Subsystem> aSubs) {
		
		// semantic checks for declaration
		// annotations of local variables
		vt.semanticChecks();
		
		// annotations of subsystem
		super.semanticChecks(vt);
		// annotations of control states
		for (IControlState s : this.controlStates()) {
			((ControlState)s).semanticChecks(vt);
		}
		// annotations of transitions
		for (ITransition t : tt) {
			((Transition)t).semanticChecks(vt);
		}
		
		for (ITransition t : tt) {
			ILabel l = t.label();
			if (l instanceof Call) {
				Call c = (Call)l;
				Subsystem callee = aSubs.get(c.calleeToken().getText());
				c.setCallee(callee);
				if (callee == null) {
					System.err.println("Semantic error at "+Common.at(c.calleeToken())+": callee '"+c.calleeToken().getText()+"' is not defined.");
					System.exit(1);
				}
				
				// separate input and output arguments of calls, infer types
				c.semantic(callee.varIn().size(), callee.varOut().size(),vt);
				
				checkCallMatch(this,callee,c);
			} else {
				Expr e = ((Expr)l);
				SemFlags s = SemFlags.createGeneral();
				e = e.semanticChecks(vt,s);
				e.checkBool();
				((Transition)t).label(e);
			}
		}
	}
	@Override
	public List<IControlState> controlStates() {
		return new LinkedList<IControlState>(states.values());
	}
}
